package io.tswf.transmutations.common.accessors

import groovy.transform.CompileStatic
import io.tswf.transmutations.common.exception.InvalidAccessorMethodException
import org.objectweb.asm.Type
import org.objectweb.asm.tree.MethodNode

@CompileStatic
class AccessorHelper {

    public static final Collection<String> GETTER_PREFIXES = [ "is", "get" ]
    public static final Collection<String> SETTER_PREFIXES = [ "set" ]

    static Type getTypeFromAccessor(MethodNode methodNode) {
        return methodNode.parameterTypes.firstOrNull() ?: methodNode.returnType
    }

    static List<String> validateSetter(MethodNode methodNode, boolean needsToThrow = true) {
        def errors = new ArrayList<String>()
        methodNode.parameterTypes.let {
            if (it.size() != 1) {
                def errorMessage = "The accessor ${methodNode.javaFullName} is invalid because setters must have single parameter!".toString()
                errors << errorMessage

                if (needsToThrow) {
                    throw new InvalidAccessorMethodException(errorMessage)
                }
            }
        }

        if (methodNode.returnType != Type.VOID_TYPE) {
            def errorMessage = "The accessor ${methodNode.javaFullName} is invalid because setters must return void!".toString()
            errors << errorMessage

            if (needsToThrow) {
                throw new InvalidAccessorMethodException(errorMessage)
            }
        }

        return errors
    }

    static List<String> validateGetter(MethodNode methodNode, boolean needsToThrow = true) {
        def errors = new ArrayList<String>()
        methodNode.parameterTypes.let {
            if (it.isNotNullOrEmpty()) {
                def errorMessage = "The accessor ${methodNode.javaFullName} is invalid because getters can not have parameters!".toString()
                errors << errorMessage

                if (needsToThrow) {
                    throw new InvalidAccessorMethodException(errorMessage)
                }
            }
        }

        if (methodNode.returnType == Type.VOID_TYPE) {
            def errorMessage = "The accessor ${methodNode.javaFullName} is invalid because getters can not return void!".toString()
            errors << errorMessage

            if (needsToThrow) {
                throw new InvalidAccessorMethodException(errorMessage)
            }
        }

        return errors
    }

    static String resolveFieldNameFromGetterMethod(MethodNode methodNode) {
        def methodName = methodNode.name
        if (methodName.startsWith("get") && methodName.length() > 3) {
            return methodName[3].toLowerCase() + methodName[4..-1]
        } else if (methodName.startsWith("is") && methodName.length() > 2) {
            return methodName[2].toLowerCase() + methodName[3..-1]
        } else {
            throw new InvalidAccessorMethodException("The accessor ${methodNode.javaFullName} is invalid because field name can't be extracted from accessor name '${methodName}' and not passed in annotation! Please, use standard 'is*' or 'get*' naming (e.g. isMyFlag or getByField) or provide additional info in annotation!")
        }
    }

    static String resolveFieldNameFromSetterMethod(MethodNode methodNode) {
        def methodName = methodNode.name
        if (methodName.startsWith("set") && methodName.length() > 3) {
            return methodName[3].toLowerCase() + methodName[4..-1]
        } else {
            throw new InvalidAccessorMethodException("The accessor ${methodNode.javaFullName} is invalid because field name can't be extracted from accessor name '${methodName}' and not passed in annotation! Please, use standard 'set*' naming (e.g. setMyFlag or setByField) or provide additional info in annotation!")
        }
    }
}
