package io.tswf.transmutations.common.reader

import groovy.transform.CompileStatic
import io.tswf.transmutations.common.exception.InvalidClassReferenceAnnotationException
import io.tswf.transmutations.common.annotation.ClassReference
import io.tswf.transmutations.common.remapping.PrimitiveClassRemapper
import org.objectweb.asm.Type
import org.objectweb.asm.tree.AnnotationNode

@CompileStatic
class ClassReferenceReader {

    public static final String CLASS_DESCRIPTOR_PARAMETER_NAME = "descriptor"
    public static final String CLASS_PARAMETER_NAME = "value"

    public static final Map<Type, Type> PRIMITIVE_MAPPERS = new HashMap<Type, Type>()
        .tap {
            put(ClassReference.BYTE.asAsmType(), byte.class.asAsmType())
            put(ClassReference.SHORT.asAsmType(), short.class.asAsmType())
            put(ClassReference.INT.asAsmType(), int.class.asAsmType())
            put(ClassReference.LONG.asAsmType(), long.class.asAsmType())
            put(ClassReference.FLOAT.asAsmType(), float.class.asAsmType())
            put(ClassReference.DOUBLE.asAsmType(), double.class.asAsmType())
            put(ClassReference.CHAR.asAsmType(), char.class.asAsmType())
            put(ClassReference.BOOLEAN.asAsmType(), boolean.class.asAsmType())
        }

    public static final Set<String> PARAMETER_NAMES = [
            CLASS_DESCRIPTOR_PARAMETER_NAME,
            CLASS_PARAMETER_NAME,
    ] as Set

    static Type getTypeFromClassRefAnnotation(AnnotationNode annotationNode) {
        findTypeFromClassRefAnnotation(annotationNode)
                ?.ensureNotNull { new InvalidClassReferenceAnnotationException("There is no class reference parameters on ${annotationNode.desc}!") }
    }

    static Type findTypeFromClassRefAnnotation(AnnotationNode annotationNode) {
        def descriptorValue = annotationNode.findValueByName(CLASS_DESCRIPTOR_PARAMETER_NAME, String)
        def classValue = annotationNode.findValueByName(CLASS_PARAMETER_NAME, Type)

        def presentedParameters = PARAMETER_NAMES
            .stream()
                .filter { annotationNode.findValueByName(it) != null }
                .toSet()

        if (presentedParameters.size() > 1) {
            throw new InvalidClassReferenceAnnotationException("The class reference annotation '${annotationNode.type.className}' must contain only one type parameter, but found ${presentedParameters} presented parameters that can not be used together")
        }

        if (presentedParameters.isEmpty()) {
            throw new InvalidClassReferenceAnnotationException("The class reference annotation '${annotationNode.type.className}' must contain only one type parameter, but found, but no one found!")
        }

        def resultType = null

        if (descriptorValue != null) {
            resultType = Type.getType(descriptorValue)
        }

        if (classValue != null) {
            resultType = classValue
        }

        resultType = PrimitiveClassRemapper.replacePrimitiveClass(resultType)

        return resultType
    }
}
